%% 
%% Theory for symbolic transition systems.
%%
%% C. Shankland and P. Robinson
%%
%% Started 2/11/00
%%
[jprop].
delete_theory_and_dependents(sts).

% Create theory
add_theory(sts).

% Inherit these theories
include_theory(stdops, share).
include_theory(iarith, all).
include_theory(localness, all).
include_theory(apart, all).
include_theory(p_list, all).
 
%% Add some extra object variables.
declare_object_vars(w).
declare_object_vars(y).
declare_object_vars(z). 

% Load the tables tactics.
require_global_tactic_file(tables).

%%%%%%%%%%%%%%%
%% Functions %%
%%%%%%%%%%%%%%%

%% The term that holds the current state info.
declare state(_,_).

%% The presicate that determines the legal states.
declare is_state(_).

%% The predicate that deals with state properties.
declare models(_,_).

%% Diamonds and boxes.
declare dia(_,_,_).
declare box(_,_,_).

%% The 'To' state for a give label
declare to(_,_,_).

%% The formula of a given label
declare formula(_,_,_).

%% For formulae within models
declare formula(_).

%% Possible transition labels
%% is_trans(State, Trans, Edge) - State is the "from" state
%% Trans is the type of transition (e.g. g transition)
%% Edge is the actual transition label (e.g. g1)
declare is_trans(_,_,_).

% Need to deal with no data

declare type_nd.
declare nd.

% axioms about no data

axiom type_nd_is_type ===  is_type(type_nd).

axiom nd_type === X:type_nd <=> X = nd.

% Axioms about models.

axiom models_iff ===
    models(State, formula(Exp)) <=> is_state(State) and Exp.

axiom models_and_iff ===
    models(State, Exp1 and Exp2) 
    <=> models(State, Exp1) and models(State, Exp2).

axiom models_or_iff ===
    models(State, Exp1 or Exp2)
    <=> models(State, Exp1) or models(State, Exp2).

axiom models_dia_iff ===
    models(State, dia(Trans, V, Form))
    <=>
    ex x is_trans(State, Trans, x)  
            and (V = V1) and formula(x, State, V1) 
            and models(to(x, State, V1), Form)
       provided x not_free_in State,
                x not_free_in Trans,
                x not_free_in Form,
                x not_free_in V,
                x not_free_in V1.

axiom models_ex_dia_iff(x:obvar) ===
    models(State, ex x:Type dia(Trans, V, Form)) 
    <=>
    ex x:Type models(State, dia(Trans, V, Form))
       provided x not_free_in Type,
                x not_free_in State,
                x not_free_in Trans.
    
axiom models_all_dia_iff(x:obvar) ===
    models(State, all x:Type dia(Trans, V, Form)) 
    <=>
    all x:Type models(State, dia(Trans, V, Form))
       provided x not_free_in Type,
                x not_free_in State,
                x not_free_in Trans.
    
axiom models_box_iff ===
    models(State, box(Trans, V, Form))
    <=>
    all x is_trans(State, Trans, x)
            and (V = V1) and formula(x, State, V1) 
             => models(to(x, State, V1), Form)
       provided x not_free_in State,
                x not_free_in Trans,
                x not_free_in Form,
                x not_free_in V,
                x not_free_in V1.

axiom models_ex_box_iff(x:obvar) ===
    models(State, ex x:Type box(Trans, V, Form)) 
    <=>
    ex x:Type models(State, box(Trans, V, Form))
       provided x not_free_in Type,
                x not_free_in State,
                x not_free_in Trans.
    
axiom models_all_box_iff(x:obvar) ===
    models(State, all x:Type box(Trans, V, Form)) 
    <=>
    all x:Type models(State, box(Trans, V, Form))
       provided x not_free_in Type,
                x not_free_in State,
                x not_free_in Trans.
    
axiom local_form(x:obvar) ===
    local x formula(x, S, L)
       provided x not_free_in S, x not_free_in L.

axiom local_models_to(x:obvar) ===
    local x models(to(x, S, L), A)
       provided x not_free_in A, x not_free_in S, x not_free_in L.

% A window opening rule for to
axiom open_to_1_eq ===
	A = B
     -----------------
      to(A, C, D) = to(B, C, D).

% No more axioms or other declarations after this point.
close_theory.

% Some theorems (really derived rules) to help with tactics
% form automated proof.

theorem open_models_1_iff ===
        A = B
      ---------------
      models(A,C) <=> models(B,C).

theorem local_and_form(x:obvar) ===
    local x F and formula(x, S, L)
       provided x not_free_in S, x not_free_in L, x not_free_in F.



theorem local_form_model(x:obvar) ===
    local x (F and formula(x, S1, L) and models(to(x, S2, L1), A))
       provided x not_free_in S1, x not_free_in S2, x not_free_in F, 
                x not_free_in L, x not_free_in L1, x not_free_in A.


require_tactic_file(wi_support).

% Save the above info in  a theory file.
% Next time Ergo is started simply "cd'ing" to the theory with
% [sts].
% will load the above info.
save.
