% Some prolog support for the sts tactics.

% Determine the appropriate rule names.

sts_names(TN, N) :-
   atom_concat(A1, N, TN),
   atom_concat(_, '.', A1).

is_trans_rule(S, L, R) :-
   sts_names(S, SName),
   sts_names(L, LName),
   concat_atom(['is_trans', SName, LName], '_', R).
to_rule(L, R) :-
   sts_names(L, LName),
   atom_concat('to_', LName, R).
formula_rule(L, R) :-
   sts_names(L, LName),
   atom_concat('formula_', LName, R).
is_state_rule(S, R) :-
   sts_names(S, SName),
   concat_atom(['is_state', SName], '_', R).
   
% Support for formula simplification
get_simplify_tac(T, ContextTerms, Tac) :-
   member(hyp = Context, ContextTerms),
   get_simplify_tac_aux(T, Context, Tac).

% General formula simplification
get_simplify_tac_aux(T, Context, Tac) :-
    T \== 'jprop.true',
    member(c(N, H), Context),
    H == T,
    !,
    Tac = trans_t(rule(assump(N))).
get_simplify_tac_aux(T, Context, Tac) :-
    T \== 'jprop.false',
    member(c(N, H), Context),
    H == 'jprop.not'(T),
    !,
    Tac = trans_ante(rule(false_iff_imp_not), 'jprop.false', [rule(assump(N))]).
get_simplify_tac_aux(T, Context, Tac) :-
    T \== 'jprop.false',
    member(c(N, H), Context),
    T == 'jprop.not'(H),
    !,
    Tac = trans_ante(rule(false_iff_not_iff), 'jprop.false', [rule(assump(N))]).
get_simplify_tac_aux('jprop.and'(_,'jprop.true'), _, Tac) :-
    !,
    Tac = trans(rule(and_true)).
get_simplify_tac_aux('jprop.and'('jprop.true',_), _, Tac) :-
    !,
    Tac = trans(rule(true_and)).
get_simplify_tac_aux('jprop.or'(_,'jprop.true'), _, Tac) :-
    !,
    Tac = trans(rule(or_true)).
get_simplify_tac_aux('jprop.or'('jprop.true',_), _, Tac) :-
    !,
    Tac = trans(rule(true_or)).
get_simplify_tac_aux('jprop.and'(_,'jprop.false'), _, Tac) :-
    !,
    Tac = trans(rule(and_false)).
get_simplify_tac_aux('jprop.and'('jprop.false',_), _, Tac) :-
    !,
    Tac = trans(rule(false_and)).
get_simplify_tac_aux('jprop.or'(_,'jprop.false'), _, Tac) :-
    !,
    Tac = trans(rule(or_false)).
get_simplify_tac_aux('jprop.or'('jprop.false',_), _, Tac) :-
    !,
    Tac = trans(rule(false_or)).
get_simplify_tac_aux('jprop.=>'(_,'jprop.true'), _, Tac) :-
    !,
    Tac = trans_t(rule(imp_true)).
get_simplify_tac_aux('jprop.=>'('jprop.true',_), _, Tac) :-
    !,
    Tac = trans(rule(true_imp)).
get_simplify_tac_aux('jprop.=>'(_,'jprop.false'), _, Tac) :-
    !,
    Tac = trans(rule(imp_false)).
get_simplify_tac_aux('jprop.=>'('jprop.false',_), _, Tac) :-
    !,
    Tac = trans(rule(false_imp)).
get_simplify_tac_aux(!!'jpred_eq_t.all' _ 'jprop.true',_, Tac) :-
    !,
    Tac = trans(rule(all_t_true), 'jprop.true').
get_simplify_tac_aux('jprop.and'(T, _), _, Tac) :-
    nonvar(T),
    T = 'weak_eq.='(X,Y),
    nonvar(X),
    var(Y),
    thawed_var(Y),
    !,
    Tac = (op([1]);trans_t(rule(eq_reflex));close_win;trans(rule(true_and))).
get_simplify_tac_aux('jprop.and'(T, _), _, Tac) :-
    nonvar(T),
    T = 'weak_eq.='(X,Y),
    X == Y,
    !,
    Tac = (op([1]);trans_t(rule(eq_reflex));close_win;trans(rule(true_and))).
get_simplify_tac_aux('weak_eq.='(X, Y), _, Tac) :-
    X == Y,
    !,
    Tac = (trans_t(rule(eq_reflex))).
get_simplify_tac_aux(!!'jpred_eq_t.ex' !x:T B,_, Tac) :-
    !x is_not_free_in B,
    ground(T),
    Tac = ( trans_rule(rule(ex_intro(!x,C))); type_instance(T,C) ).
% Rewrite is_state...
get_simplify_tac_aux(Term, _, Tac) :-
    Term = 'jprop.and'('sts.is_state'(_(S,_)),_),
    nonvar(S),
    !,
    is_state_rule(S, SR),
    Tac = (op([1]);trans_t(rule(SR));close_win;trans(rule(true_and))).
get_simplify_tac_aux(Term, _, Tac) :-
    Term = 'sts.is_state'(_(S,_)),
    nonvar(S),
    !,
    is_state_rule(S, SR),
    Tac = (trans_t(rule(SR))).
% Rewrite is_trans formulae in ex
get_simplify_tac_aux(Term, _, Tac) :-
    Term = !!_ _ _and(B, _),
    nonvar(B),
    B = 'sts.is_trans'(_state(S,_),L,_),
    bound_var(Term, [!x1]),
    is_trans_rule(S,L,R),
    !,
    Tac = ( op([b(!x1),1]); trans(rule(R)); close_win ).
% Rewrite is_trans formulae in all
get_simplify_tac_aux(Term, _, Tac) :-
    Term = !!_ _ _imp(_and(B, _),_),
    nonvar(B),
    B = 'sts.is_trans'(_state(S,_),L,_),
    bound_var(Term, [!x1]),
    is_trans_rule(S,L,R),
    !,
    Tac = ( op([b(!x1),1,1]); trans(rule(R)); close_win ).
% Rewrite models
get_simplify_tac_aux(Term, _, Tac) :-
    Term = 'sts.models'(_,D),
    nonvar(D),
    !,
    get_simplify_tac_models(D, Tac).
% Rewrite ex to an instance
get_simplify_tac_aux(Term, _, Tac) :-
    quant(Term), 
    substitute(Term, _, Term1),
    quantify(Term1, 'jpred_eq_t.ex', [_(BV,_)], _),
    !,
    Tac = ( trans_rule(rule(ex_intro(BV,_))); discharge_type ).
% Support for "iterating" in ex
get_simplify_tac_aux(Term, _, Tac) :-
    quant(Term), 
    substitute(Term, _, Term1),
    quantify(Term1, 'jpred.ex', [!x], 'jprop.and'(A1,_)),
    nonvar(A1),
    ( A1 = 'jprop.false'
      ->
        Tac = (op([b(!x)]);trans(rule(false_and));close_win;trans(rule(ex_false)))
      ;
      A1 = 'jprop.or'(X1,_)
      ->
        nonvar(X1), X1 = 'weak_eq.='(_,_),!,
        Tac = ( trans_ante(rule(one_pt_ex_or(!x)), [rule(local_form_model)]);
                dia_tac_aux_2)
      ;
        A1 = 'weak_eq.='(_,_),!,
        Tac = (trans_ante(rule(one_pt_ex(!x)), _, [rule(local_form_model)]);
                dia_tac_aux_1)
    ).
% Support for "iterating" in all
get_simplify_tac_aux(Term, _, Tac) :-
    quant(Term), 
    substitute(Term, _, Term1),
    quantify(Term1, 'jpred.all', [!x], 'jprop.=>'('jprop.and'(A1,_),_)),
    nonvar(A1),
    ( A1 = 'jprop.false'
      ->
        Tac = (op([b(!x)]);op([1]);trans(rule(false_and));
               close_win;trans(rule(false_imp));close_win;trans(rule(all_true)))
      ;
      A1 = 'jprop.or'(X1,_)
      ->
        nonvar(X1), X1 = 'weak_eq.='(_,_),!,
        Tac = ( trans_ante(rule(one_pt_all_and_imp(!x)), 
                           [rule(local_and_form), rule(local_models_to)]);
                box_tac_aux_2)
      ;
        A1 = 'weak_eq.='(_,_),!,
        Tac = (trans_ante(rule(one_pt_all_imp(!x)), [rule(local_and_form), rule(local_models_to)]);
                box_tac_aux_1)
    ).
% IF nothing else works try rewriting models in second subformula
get_simplify_tac_aux(Term, _, Tac) :-
    nonvar(Term), 
    Term = _(_,M), 
    nonvar(M),
    M = 'sts.models'(_,_),
    !,
    Tac = op([2]).
% IF nothing else works try rewriting the body of a quantifier
get_simplify_tac_aux(Term, _, Tac) :-
    nonvar(Term), 
    Term = _(_,M), 
    quant(M), 
    substitute(M, _, M1),
    quantify(M1, _, _, B),
    ( B = _(_,_(_,_(B1,_))) ; B = _(_(_,_(_,B1)),_) ), 
    nonvar(B1), 
    B1 = 'sts.formula'(_,_,_),
    !,
    Tac = op([2]).

% Rewrite models(S, all x:T dia(...))
get_simplify_tac_models(D, Tac) :-
    quant(D),
    substitute(D, _, D1),
    quantify(D1, 'jpred_eq_t.all', [_(XDA,_)], Body),
    nonvar(Body),
    Body = 'sts.dia'(_,_,_),
    !,
    Tac = (trans(rule(models_all_dia_iff(XDA)));op([b(XDA)])).
% Rewrite models(S, all x:T box(...))
get_simplify_tac_models(D, Tac) :-
    quant(D),
    substitute(D, _, D1),
    quantify(D1, 'jpred_eq_t.all', [_(XDA,_)], Body),
    nonvar(Body),
    Body = 'sts.box'(_,_,_),
    !,
    Tac = (trans(rule(models_all_box_iff(XDA)));op([b(XDA)])).
% Rewrite models(S, ex x:T dia(...))
get_simplify_tac_models(D, Tac) :-
    quant(D),
    substitute(D, _, D1),
    quantify(D1, 'jpred_eq_t.ex', [_(XDA,_)], Body),
    nonvar(Body),
    Body = 'sts.dia'(_,_,_),
    !,
    Tac = (trans(rule(models_ex_dia_iff(XDA)));op([b(XDA)])).
% Rewrite models(S, ex x:T box(...))
get_simplify_tac_models(D, Tac) :-
    quant(D),
    substitute(D, _, D1),
    quantify(D1, 'jpred_eq_t.ex', [_(XDA,_)], Body),
    nonvar(Body),
    Body = 'sts.box'(_,_,_),
    !,
    Tac = (trans(rule(models_ex_box_iff(XDA)));op([b(XDA)])).
% Rewrite models(S, formula(...))
get_simplify_tac_models('sts.formula'(_), Tac) :-
    !,
    Tac = trans(rule(models_iff)).
% Rewrite models(S, dia(...))
get_simplify_tac_models('sts.dia'(_,_,_), Tac) :-
    !,
    Tac = trans(rule(models_dia_iff)).
% Rewrite models(S, box(...))
get_simplify_tac_models('sts.box'(_,_,_), Tac) :-
    !,
    Tac = trans(rule(models_box_iff)).
% Rewrite models(S, or(...))
get_simplify_tac_models('jprop.or'(_,_), Tac) :-
    !,
    Tac = (trans(rule(models_or_iff));op([1])).
% Rewrite models(S, and(...))
get_simplify_tac_models('jprop.and'(_,_), Tac) :-
    !,
    Tac = (trans(rule(models_and_iff));op([1])).

% Some support for arithmetic simplification

simp_arith(T, ContextTerms, Tac) :-
fail.
%   member((hyp=Context), ContextTerms),
%   nonvar(T),
%   T = _WR(T1,_),
%   nonvar(T1),
%   T1 = OP(A1,A2),
%   operator_name(OP, ON),
%   simp_arith2(A1, A2, ON, Context, Tac).


simp_arith2(A1, A2, ON, _, Tac) :-
   'iarith:is_int_const'(A1), 
   'iarith:is_int_const'(A2),
   !,
   ( ON = eq
     -> R = do_arith_eq
     ;
     ON = lt
     -> R = do_arith_lt
     ;
     ON = le, R = do_arith_le
   ),
   Tac = arith_simp1(R).
     
simp_arith2(A1, A2, ON, Context, Tac) :-
   member(c(ID, H), Context),
   nonvar(H),
   H = HOP(HA1,HA2),
   operator_name(HOP, HN),
   simp_arith3(A1, A2, ON, HA1, HA2, HN, ID, Context, Tac).

% eq_lt_lt1 ===
%    A = Int2 => (A < Int1 <=> Int2 < Int1)
simp_arith3(A1, A2, lt, HA1, HA2, HN, ID, _, Tac) :-
   'iarith:int_const_to_int'(A2, A2N),
   'iarith:int_const_to_int'(HA2, HA2N),
   A2N =< HA2N,
   A1 == HA1,
   HN = eq,
   !,
   Tac = arith_simp2(eq_lt_lt1, do_arith_lt, ID).
% eq_lt_lt2 ===
%    Int2 = A => (A < Int1 <=> Int2 < Int1)
simp_arith3(A1, A2, lt, HA1, HA2, HN, ID, _, Tac) :-
   'iarith:int_const_to_int'(A2,A2N),
   'iarith:int_const_to_int'(HA1, HA1N),
   A2N =< HA1N,
   A1 == HA2,
   HN = eq,
   !,
   Tac = arith_simp2(eq_lt_lt2, do_arith_lt, ID).
% eq_lt_lt3 ===
%    A = Int2 => (Int1 < A <=> Int1 < Int2)
simp_arith3(A1, A2, lt, HA1, HA2, HN, ID, _, Tac) :-
   'iarith:int_const_to_int'(A1, A1N),
   'iarith:int_const_to_int'(HA2, HA2N),
   HA2N =< A1N,
   A2 == HA1,
   HN = eq,
   !,
   Tac = arith_simp2(eq_lt_lt3, do_arith_lt, ID).
% eq_lt_lt4 ===
%    Int2 = A => (Int1 < A <=> Int1 < Int2)
simp_arith3(A1, A2, lt, HA1, HA2, HN, ID, _, Tac) :-
   'iarith:int_const_to_int'(A1, A1N),
   'iarith:int_const_to_int'(HA1, HA1N),
   HA1N =< A1N,
   A2 == HA2,
   HN = eq,
   !,
   Tac = arith_simp2(eq_lt_lt4, do_arith_lt, ID).

% eq_le_le1 ===
%    A = Int2 => (A =< Int1 <=> Int2 =< Int1)
simp_arith3(A1, A2, le, HA1, HA2, HN, ID, _, Tac) :-
   'iarith:int_const_to_int'(A2, A2N),
   'iarith:int_const_to_int'(HA2, HA2N),
   HA2N > A2N,
   A1 == HA1,
   HN = eq,
   !,
   Tac = arith_simp2(eq_le_le1, do_arith_le, ID).
% eq_le_le2 ===
%    Int2 = A => (A =< Int1 <=> Int2 =< Int1)
simp_arith3(A1, A2, le, HA1, HA2, HN, ID, _, Tac) :-
   'iarith:int_const_to_int'(A2, A2N),
   'iarith:int_const_to_int'(HA1, HA1N),
   HA1N > A2N,
   A1 == HA2,
   HN = eq,
   !,
   Tac = arith_simp2(eq_le_le2, do_arith_le, ID).
% eq_le_le3 ===
%    A = Int2 => (Int1 =< A <=> Int1 =< Int2)
simp_arith3(A1, A2, le, HA1, HA2, HN, ID, _, Tac) :-
   'iarith:int_const_to_int'(A1, A1N),
   'iarith:int_const_to_int'(HA2, HA2N),
   HA2N < A1N,
   A2 == HA1,
   HN = eq,
   !,
   Tac = arith_simp2(eq_le_le3, do_arith_le, ID).
% eq_le_le4 ===
%    Int2 = A => (Int1 =< A <=> Int1 =< Int2)
simp_arith3(A1, A2, le, HA1, HA2, HN, ID, _, Tac) :-
   'iarith:int_const_to_int'(A1, A1N),
   'iarith:int_const_to_int'(HA1, HA1N),
   HA1N < A1N,
   A2 == HA2,
   HN = eq,
   !,
   Tac = arith_simp2(eq_le_le4, do_arith_le, ID).

% eq_imp_eq_eq1 ===
%     A = Int2 => (A = Int1 <=> Int2 = Int1)
simp_arith3(A1, A2, eq, HA1, HA2, HN, ID, _, Tac) :-
   'iarith:int_const_to_int'(A2, A2N),
   'iarith:int_const_to_int'(HA2, HA2N),
   A2N \= HA2N,
   A1 == HA1,
   HN = eq,
   !,
   Tac = arith_simp2(eq_imp_eq_eq1, do_arith_eq, ID).
% eq_imp_eq_eq2 ===
%     Int2 = A => (A = Int1 <=> Int2 = Int1)
simp_arith3(A1, A2, eq, HA1, HA2, HN, ID, _, Tac) :-
   'iarith:int_const_to_int'(A2, A2N),
   'iarith:int_const_to_int'(HA1, HA1N),
   A2N \= HA1N,
   A1 == HA2,
   HN = eq,
   !,
   Tac = arith_simp2(eq_imp_eq_eq2, do_arith_eq, ID).
% eq_imp_eq_eq3 ===
%     A = Int2 => (Int1 = A <=> Int2 = Int1)
simp_arith3(A1, A2, eq, HA1, HA2, HN, ID, _, Tac) :-
   'iarith:int_const_to_int'(A1, A1N),
   'iarith:int_const_to_int'(HA2, HA2N),
   A1N \= HA2N,
   A2 == HA1,
   HN = eq,
   !,
   Tac = arith_simp2(eq_imp_eq_eq3, do_arith_eq, ID).
% eq_imp_eq_eq4 ===
%     Int2 = A => (Int1 = A <=> Int2 = Int1)
simp_arith3(A1, A2, eq, HA1, HA2, HN, ID, _, Tac) :-
   'iarith:int_const_to_int'(A1, A1N),
   'iarith:int_const_to_int'(HA1, HA1N),
   A1N \= HA1N,
   A2 == HA2,
   HN = eq,
   !,
   Tac = arith_simp2(eq_imp_eq_eq4, do_arith_eq, ID).

% lt_eq_false1 ===
% A:ints and Int2:ints and Int1:ints and A < Int2 and Int2 =< Int1
%      => (A = Int1 <=> false)
simp_arith3(A1, A2, eq, HA1, HA2, HN, ID, Context, Tac) :-
   'iarith:int_const_to_int'(A2, A2N),
   'iarith:int_const_to_int'(HA2, HA2N),
   HA2N =< A2N,
   A1 == HA1,
   HN = lt,
   member(c(ID1, 'types.:'(XA1, 'iarith.ints')), Context),
   XA1 == A1,
   !,
   Tac = arith_simp3(lt_eq_false1, ID, ID1).
% lt_eq_false2 ===
% Int1:ints and Int2:ints and Int2:ints and Int1 < Int2 and Int2 =< C
%      => (C = Int1 <=> false)
simp_arith3(A1, A2, eq, HA1, HA2, HN, ID, Context, Tac) :-
   'iarith:int_const_to_int'(A2, A2N),
   'iarith:int_const_to_int'(HA1, HA1N),
   HA1N > A2N,
   A1 == HA2,
   HN = le,
   member(c(ID1, 'types.:'(XA1, 'iarith.ints')), Context),
   XA1 == A1,
   !,
   Tac = arith_simp4(lt_eq_false2, ID, ID1).
% lt_eq_false3
% Int2:ints and B:ints and Int1:ints and Int2 < B and Int1 =< Int2
%      => (B = Int1 <=> false)
simp_arith3(A1, A2, eq, HA1, HA2, HN, ID, Context, Tac) :-
   'iarith:int_const_to_int'(A2, A2N),
   'iarith:int_const_to_int'(HA1, HA1N),
   A2N =< HA1N,
   A1 == HA2,
   HN = lt,
   member(c(ID1, 'types.:'(XA1, 'iarith.ints')), Context),
   XA1 == A1,
   !,
   Tac = arith_simp5(lt_eq_false3, ID, ID1).
% lt_eq_false4 ===
% Int2:ints and Int1:ints and C:ints and Int2 < Int1 and C =< Int2
%      => (C = Int1 <=> false)
simp_arith3(A1, A2, eq, HA1, HA2, HN, ID, Context, Tac) :-
   'iarith:int_const_to_int'(A2, A2N),
   'iarith:int_const_to_int'(HA2, HA2N),
   A2N > HA2N,
   A1 == HA1,
   HN = le,
   member(c(ID1, 'types.:'(XA1, 'iarith.ints')), Context),
   XA1 == A1,
   !,
   Tac = arith_simp4(lt_eq_false4, ID, ID1).

% lt_eq_false1 ===
% Int1:ints and Int2:ints and C:ints and Int1 < Int2 and Int2 =< C
%      => (Int1 = C <=> false)
simp_arith3(A1, A2, eq, HA1, HA2, HN, ID, Context, Tac) :-
   'iarith:int_const_to_int'(A1, A1N),
   'iarith:int_const_to_int'(HA1, HA1N),
   HA1N > A1N,
   A2 == HA2,
   HN = le,
   member(c(ID1, 'types.:'(XA2, 'iarith.ints')), Context),
   XA2 == A2,
   !,
   Tac = arith_simp4(lt_eq_false1, ID, ID1).
% lt_eq_false2 ===
%    A:ints and Int2:ints and Int1:ints and A < Int2 and Int2 =< Int1
%      => (Int1 = A <=> false)
simp_arith3(A1, A2, eq, HA1, HA2, HN, ID, Context, Tac) :-
   'iarith:int_const_to_int'(A1, A1N),
   'iarith:int_const_to_int'(HA2, HA2N),
   HA2N =< A1N,
   A2 == HA1,
   HN = lt,
   member(c(ID1, 'types.:'(XA2, 'iarith.ints')), Context),
   XA2 == A2,
   !,
   Tac = arith_simp3(lt_eq_false2, ID, ID1).
% lt_eq_false3 ===
%    Int2:ints and Int1:ints and C:ints and Int2 < Int1 and C =< Int2
%      => (Int1 = C <=> false)
simp_arith3(A1, A2, eq, HA1, HA2, HN, ID, Context, Tac) :-
   'iarith:int_const_to_int'(A1, A1N),
   'iarith:int_const_to_int'(HA2, HA2N),
   A1N > HA2N,
   A2 == HA1,
   HN = le,
   member(c(ID1, 'types.:'(XA2, 'iarith.ints')), Context),
   XA2 == A2,
   !,
   Tac = arith_simp4(lt_eq_false3, ID, ID1).
% lt_eq_false4 ===
%    Int2:ints and B:ints and Int1:ints and Int2 < B and Int1 =< Int2
%      => (Int1 = B <=> false)
simp_arith3(A1, A2, eq, HA1, HA2, HN, ID, Context, Tac) :-
   'iarith:int_const_to_int'(A1, A1N),
   'iarith:int_const_to_int'(HA1, HA1N),
   A1N =< HA1N,
   A2 == HA2,
   HN = lt,
   member(c(ID1, 'types.:'(XA2, 'iarith.ints')), Context),
   XA2 == A2,
   !,
   Tac = arith_simp5(lt_eq_false4, ID, ID1).
   
% lt_lt_false ===
%    Int:ints and B:ints and B < Int => (Int < B <=> false)
simp_arith3(A1, A2, lt, HA1, HA2, HN, ID, Context, Tac) :-
   'iarith:is_int_const'(A1),
   A2 == HA1,
   A1 == HA2,
   HN = lt,
   member(c(ID1, 'types.:'(XA2, 'iarith.ints')), Context),
   XA2 == A2,
   !,
   Tac = arith_simp6(ID, ID1).
% lt_lt_false ===
%    A:ints and Int:ints and Int < A => (A < Int <=> false)
simp_arith3(A1, A2, lt, HA1, HA2, HN, ID, Context, Tac) :-
   'iarith:is_int_const'(A2),
   A2 == HA1,
   A1 == HA2,
   HN = lt,
   member(c(ID1, 'types.:'(XA1, 'iarith.ints')), Context),
   XA1 == A1,
   !,
   Tac = arith_simp7(ID, ID1).
   
   
